@@ -21,12 +21,12 @@
  * Boston, MA 02111-1307, USA.
  */
 
-#include <config.h>
 #include <gmodule.h>
 #include <libgda/gda-client.h>
 #include <libgda/gda-config.h>
 #include <libgda/gda-intl.h>
 #include <libgda/gda-log.h>
+#include <libgda/gda-server-provider.h>
 #include <string.h>
 #include "gda-marshal.h"
 
@@ -52,8 +52,10 @@ static void gda_client_class_init (GdaClientClass *klass);
 static void gda_client_init       (GdaClient *client, GdaClientClass *klass);
 static void gda_client_finalize   (GObject *object);
 
+static void connection_error_cb (GdaConnection *cnc, GList *error_list, gpointer user_data);
+
 enum {
-	ERROR,
+	EVENT_NOTIFICATION,
 	LAST_SIGNAL
 };
 
@@ -77,71 +79,15 @@ emit_client_error (GdaClient *client, GdaConnection *cnc, const gchar *format, .
 	vsprintf (sz, format, args);
 	va_end (args);
 
+	g_print ("Error: %s\n", sz);
+
 	/* create the error list */
 	error = gda_error_new ();
 	gda_error_set_description (error, sz);
 	gda_error_set_source (error, "[GDA client]");
 
 	list = g_list_append (NULL, error);
-	g_signal_emit (G_OBJECT (client), gda_client_signals[ERROR], 0, cnc, list);
-}
-
-/*
- * Callbacks
- */
-
-static void
-connection_error_cb (GdaConnection *cnc, GList *error_list, gpointer user_data)
-{
-	GdaClient *client = (GdaClient *) user_data;
-
-	g_return_if_fail (GDA_IS_CLIENT (client));
-	g_signal_emit (G_OBJECT (client), gda_client_signals[ERROR], 0, cnc, error_list);
-}
-
-static void
-cnc_weak_cb (gpointer user_data, GObject *object)
-{
-	GdaConnection *cnc = (GdaConnection *) object;
-	GdaClient *client = (GdaClient *) user_data;
-
-	g_return_if_fail (GDA_IS_CONNECTION (cnc));
-	g_return_if_fail (GDA_IS_CLIENT (client));
-
-	client->priv->connections = g_list_remove (client->priv->connections, cnc);
-}
-
-/*
- * GdaClient class implementation
- */
-
-static void
-gda_client_class_init (GdaClientClass *klass)
-{
-	GObjectClass *object_class = G_OBJECT_CLASS (klass);
-
-	parent_class = g_type_class_peek_parent (klass);
-
-	gda_client_signals[ERROR] =
-		g_signal_new ("error",
-			      G_TYPE_FROM_CLASS (object_class),
-			      G_SIGNAL_RUN_LAST,
-			      G_STRUCT_OFFSET (GdaClientClass, error),
-			      NULL, NULL,
-			      gda_marshal_VOID__POINTER_POINTER,
-			      G_TYPE_NONE, 2, G_TYPE_POINTER, G_TYPE_POINTER);
-
-	object_class->finalize = gda_client_finalize;
-}
-
-static void
-gda_client_init (GdaClient *client, GdaClientClass *klass)
-{
-	g_return_if_fail (GDA_IS_CLIENT (client));
-
-	client->priv = g_new0 (GdaClientPrivate, 1);
-	client->priv->providers = g_hash_table_new (g_str_hash, g_str_equal);
-	client->priv->connections = NULL;
+	connection_error_cb (cnc, list, client);
 }
 
 static void
@@ -158,6 +104,113 @@ free_hash_provider (gpointer key, gpointer value, gpointer user_data)
 	}
 }
 
+/*
+ * Callbacks
+ */
+
+static void
+connection_error_cb (GdaConnection *cnc, GList *error_list, gpointer user_data)
+{
+	GList *l;
+	GdaClient *client = (GdaClient *) user_data;
+
+	g_return_if_fail (GDA_IS_CLIENT (client));
+
+	/* notify actions */
+	for (l = error_list; l != NULL; l = l->next)
+		gda_client_notify_error_event (client, cnc, GDA_ERROR (l->data));
+}
+
+static void
+cnc_weak_cb (gpointer user_data, GObject *object)
+{
+	GdaConnection *cnc = (GdaConnection *) object;
+	GdaClient *client = (GdaClient *) user_data;
+
+	g_return_if_fail (GDA_IS_CONNECTION (cnc));
+	g_return_if_fail (GDA_IS_CLIENT (client));
+
+	client->priv->connections = g_list_remove (client->priv->connections, cnc);
+}
+
+typedef struct {
+	GdaClient *client;
+	GdaServerProvider *provider;
+	gboolean already_removed;
+} prv_weak_cb_data;
+
+static gboolean
+remove_provider_in_hash (gpointer key, gpointer value, gpointer user_data)
+{
+	LoadedProvider *prv = value;
+	prv_weak_cb_data *cb_data = user_data;
+
+	if (prv->provider == cb_data->provider && !cb_data->already_removed) {
+		g_free (key);
+		g_module_close (prv->handle);
+		g_free (prv);
+		cb_data->already_removed = TRUE;
+	}
+	return TRUE;
+}
+
+static void
+provider_weak_cb (gpointer user_data, GObject *object)
+{
+	prv_weak_cb_data cb_data;
+	GdaServerProvider *provider = (GdaServerProvider *) object;
+	GdaClient *client = (GdaClient *) user_data;
+
+	g_return_if_fail (GDA_IS_SERVER_PROVIDER (provider));
+	g_return_if_fail (GDA_IS_CLIENT (client));
+
+	cb_data.client = client;
+	cb_data.provider = provider;
+	cb_data.already_removed = FALSE;
+	g_hash_table_foreach_remove (client->priv->providers, (GHRFunc) remove_provider_in_hash, &cb_data);
+}
+
+/*
+ * GdaClient class implementation
+ */
+
+static void
+gda_client_class_init (GdaClientClass *klass)
+{
+	GObjectClass *object_class = G_OBJECT_CLASS (klass);
+
+	parent_class = g_type_class_peek_parent (klass);
+
+	gda_client_signals[EVENT_NOTIFICATION] =
+		g_signal_new ("event_notification",
+			      G_TYPE_FROM_CLASS (object_class),
+			      G_SIGNAL_RUN_LAST,
+			      G_STRUCT_OFFSET (GdaClientClass, event_notification),
+			      NULL, NULL,
+			      gda_marshal_VOID__POINTER_INT_POINTER,
+			      G_TYPE_NONE, 3, G_TYPE_POINTER, G_TYPE_INT, G_TYPE_POINTER);
+
+	object_class->finalize = gda_client_finalize;
+}
+
+static void
+gda_client_init (GdaClient *client, GdaClientClass *klass)
+{
+	g_return_if_fail (GDA_IS_CLIENT (client));
+
+	client->priv = g_new0 (GdaClientPrivate, 1);
+	client->priv->providers = g_hash_table_new (g_str_hash, g_str_equal);
+	client->priv->connections = NULL;
+}
+
+static void
+remove_weak_ref (gpointer key, gpointer value, gpointer user_data)
+{
+	LoadedProvider *prv = (LoadedProvider *) value;
+
+	g_object_weak_unref (G_OBJECT (prv->provider), (GWeakNotify) provider_weak_cb, G_OBJECT (user_data));
+}
+
 static void
 gda_client_finalize (GObject *object)
 {
@@ -168,6 +221,7 @@ gda_client_finalize (GObject *object)
 	/* free memory */
 	gda_client_close_all_connections (client);
 
+	g_hash_table_foreach (client->priv->providers, (GHFunc) remove_weak_ref, client);
 	g_hash_table_foreach (client->priv->providers, (GHFunc) free_hash_provider, NULL);
 	g_hash_table_destroy (client->priv->providers);
 	client->priv->providers = NULL;
@@ -226,6 +280,7 @@ gda_client_new (void)
  * @dsn: data source name.
  * @username: user name.
  * @password: password for @username.
+ * @options: Options for the connection.
  *
  * Establish a connection to a data source.
  *
@@ -239,7 +294,8 @@ GdaConnection *
 gda_client_open_connection (GdaClient *client,
 			    const gchar *dsn,
 			    const gchar *username,
-			    const gchar *password)
+			    const gchar *password,
+			    GdaConnectionOptions options)
 {
 	GdaConnection *cnc;
 	LoadedProvider *prv;
@@ -258,6 +314,7 @@ gda_client_open_connection (GdaClient *client,
 	cnc = gda_client_find_connection (client, dsn, username, password);
 	if (cnc) {
 		g_object_ref (G_OBJECT (cnc));
+		gda_client_notify_connection_opened_event (client, cnc);
 		return cnc;
 	}
 
@@ -290,6 +347,8 @@ gda_client_open_connection (GdaClient *client,
 			return NULL;
 		}
 
+		g_module_make_resident (prv->handle);
+
 		g_module_symbol (prv->handle, "plugin_get_name",
 				 (gpointer) &prv->plugin_get_name);
 		g_module_symbol (prv->handle, "plugin_get_description",
@@ -318,12 +377,13 @@ gda_client_open_connection (GdaClient *client,
 		}
 
 		g_object_ref (G_OBJECT (prv->provider));
+		g_object_weak_ref (G_OBJECT (prv->provider), (GWeakNotify) provider_weak_cb, client);
 		g_hash_table_insert (client->priv->providers,
 				     g_strdup (dsn_info->provider),
 				     prv);
 	}
 
-	cnc = gda_connection_new (client, prv->provider, dsn, username, password);
+	cnc = gda_connection_new (client, prv->provider, dsn, username, password, options);
 
 	if (!GDA_IS_CONNECTION (cnc)) {
 		gda_config_free_data_source_info (dsn_info);
@@ -360,7 +420,8 @@ gda_client_open_connection (GdaClient *client,
 GdaConnection *
 gda_client_open_connection_from_string (GdaClient *client,
 					const gchar *provider_id,
-					const gchar *cnc_string)
+					const gchar *cnc_string,
+					GdaConnectionOptions options)
 {
 	GdaDataSourceInfo *dsn_info;
 	GdaConnection *cnc;
@@ -388,7 +449,8 @@ gda_client_open_connection_from_string (GdaClient *client,
 	/* open the connection */
 	cnc = gda_client_open_connection (client, dsn_info->name,
 					  dsn_info->username,
-					  dsn_info->password);
+					  dsn_info->password,
+					  options);
 
 	/* free memory */
 	gda_config_remove_data_source (dsn_info->name);
@@ -487,6 +549,156 @@ gda_client_close_all_connections (GdaClient *client)
 	client->priv->connections = NULL;
 }
 
+/**
+ * gda_client_notify_event
+ * @client: a #GdaClient object.
+ * @cnc: a #GdaConnection object where the event has occurred.
+ * @event: event ID.
+ * @params: parameters associated with the event.
+ *
+ * Notify an event to the given #GdaClient's listeners. The event can be
+ * anything (see #GdaClientEvent) ranging from a connection opening
+ * operation, to changes made to a table in an underlying database.
+ */
+void
+gda_client_notify_event (GdaClient *client,
+			 GdaConnection *cnc,
+			 GdaClientEvent event,
+			 GdaParameterList *params)
+{
+	g_return_if_fail (GDA_IS_CLIENT (client));
+	g_signal_emit (G_OBJECT (client), gda_client_signals[EVENT_NOTIFICATION], 0,
+		       cnc, event, params);
+}
+
+/**
+ * gda_client_notify_error_event
+ * @client: a #GdaClient object.
+ * @cnc: a #GdaConnection object.
+ * @error: the error to be notified.
+ *
+ * Notify the given #GdaClient of the #GDA_CLIENT_EVENT_ERROR event.
+ */
+void
+gda_client_notify_error_event (GdaClient *client, GdaConnection *cnc, GdaError *error)
+{
+	GdaParameterList *params;
+
+	g_return_if_fail (GDA_IS_CLIENT (client));
+	g_return_if_fail (GDA_IS_CONNECTION (cnc));
+	g_return_if_fail (error != NULL);
+
+	params = gda_parameter_list_new ();
+	gda_parameter_list_add_parameter (params, gda_parameter_new_gobject ("error", error));
+	gda_client_notify_event (client, cnc, GDA_CLIENT_EVENT_ERROR, params);
+
+	gda_parameter_list_free (params);
+}
+
+/**
+ * gda_client_notify_connection_opened_event
+ * @client: a #GdaClient object.
+ * @cnc: a #GdaConnection object.
+ *
+ * Notify the given #GdaClient of the #GDA_CLIENT_EVENT_CONNECTION_OPENED event.
+ */
+void
+gda_client_notify_connection_opened_event (GdaClient *client, GdaConnection *cnc)
+{
+	g_return_if_fail (GDA_IS_CLIENT (client));
+	g_return_if_fail (GDA_IS_CONNECTION (cnc));
+
+	gda_client_notify_event (client, cnc, GDA_CLIENT_EVENT_CONNECTION_OPENED, NULL);
+}
+
+/**
+ * gda_client_notify_connection_closed_event
+ * @client: a #GdaClient object.
+ * @cnc: a #GdaConnection object.
+ *
+ * Notify the given #GdaClient of the #GDA_CLIENT_EVENT_CONNECTION_CLOSED event.
+ */
+void
+gda_client_notify_connection_closed_event (GdaClient *client, GdaConnection *cnc)
+{
+	g_return_if_fail (GDA_IS_CLIENT (client));
+	g_return_if_fail (GDA_IS_CONNECTION (cnc));
+
+	gda_client_notify_event (client, cnc, GDA_CLIENT_EVENT_CONNECTION_CLOSED, NULL);
+}
+
+/**
+ * gda_client_notify_transaction_started_event
+ * @client: a #GdaClient object.
+ * @cnc: a #GdaConnection object.
+ * @xaction: a #GdaTransaction object.
+ *
+ * Notify the given #GdaClient of the #GDA_CLIENT_EVENT_TRANSACTION_STARTED event.
+ */
+void
+gda_client_notify_transaction_started_event (GdaClient *client, GdaConnection *cnc, GdaTransaction *xaction)
+{
+	GdaParameterList *params;
+
+	g_return_if_fail (GDA_IS_CLIENT (client));
+	g_return_if_fail (GDA_IS_CONNECTION (cnc));
+	g_return_if_fail (GDA_IS_TRANSACTION (xaction));
+
+	params = gda_parameter_list_new ();
+	gda_parameter_list_add_parameter (params, gda_parameter_new_gobject ("transaction", xaction));
+	gda_client_notify_event (client, cnc, GDA_CLIENT_EVENT_TRANSACTION_STARTED, params);
+
+	gda_parameter_list_free (params);
+}
+
+/**
+ * gda_client_notify_transaction_committed_event
+ * @client: a #GdaClient object.
+ * @cnc: a #GdaConnection object.
+ * @xaction: a #GdaTransaction object.
+ *
+ * Notify the given #GdaClient of the #GDA_CLIENT_EVENT_TRANSACTION_COMMITTED event.
+ */
+void
+gda_client_notify_transaction_committed_event (GdaClient *client, GdaConnection *cnc, GdaTransaction *xaction)
+{
+	GdaParameterList *params;
+
+	g_return_if_fail (GDA_IS_CLIENT (client));
+	g_return_if_fail (GDA_IS_CONNECTION (cnc));
+	g_return_if_fail (GDA_IS_TRANSACTION (xaction));
+
+	params = gda_parameter_list_new ();
+	gda_parameter_list_add_parameter (params, gda_parameter_new_gobject ("transaction", xaction));
+	gda_client_notify_event (client, cnc, GDA_CLIENT_EVENT_TRANSACTION_COMMITTED, params);
+
+	gda_parameter_list_free (params);
+}
+
+/**
+ * gda_client_notify_transaction_cancelled_event
+ * @client: a #GdaClient object.
+ * @cnc: a #GdaConnection object.
+ * @xaction: a #GdaTransaction object.
+ *
+ * Notify the given #GdaClient of the #GDA_CLIENT_EVENT_TRANSACTION_CANCELLED event.
+ */
+void
+gda_client_notify_transaction_cancelled_event (GdaClient *client, GdaConnection *cnc, GdaTransaction *xaction)
+{
+	GdaParameterList *params;
+
+	g_return_if_fail (GDA_IS_CLIENT (client));
+	g_return_if_fail (GDA_IS_CONNECTION (cnc));
+	g_return_if_fail (GDA_IS_TRANSACTION (xaction));
+
+	params = gda_parameter_list_new ();
+	gda_parameter_list_add_parameter (params, gda_parameter_new_gobject ("transaction", xaction));
+	gda_client_notify_event (client, cnc, GDA_CLIENT_EVENT_TRANSACTION_CANCELLED, params);
+
+	gda_parameter_list_free (params);
+}
+
 /**
  * gda_client_begin_transaction
  * @client: a #GdaClient object.
