@@ -67,7 +67,7 @@ display_recordset_data (GdaDataModel *model)
 	GdaCommandType cmd_type;
 	const gchar *cmd_text, *type_text;
 
-	g_return_if_fail (GDA_IS_RECORDSET (model));
+	g_return_if_fail (GDA_IS_DATA_MODEL (model));
 
 	cols = gda_data_model_get_n_columns (model);
 	rows = gda_data_model_get_n_rows (model);
@@ -75,28 +75,62 @@ display_recordset_data (GdaDataModel *model)
 	g_print ("Displaying recordset %p, with %d columns and %d rows:\n",
 		 model, cols, rows);
 
-	cmd_type = gda_recordset_get_command_type (GDA_RECORDSET (model));
+	cmd_type = gda_data_model_get_command_type (model);
 	type_text = command_type_to_string (cmd_type);
-	cmd_text = gda_recordset_get_command_text (GDA_RECORDSET (model));
+	cmd_text = gda_data_model_get_command_text (model);
 	g_print ("The query (type '%s') leading up to this result was:\n\t'%s'\n\n",
 			type_text, cmd_text ? cmd_text : "<null>");
 
 	for (r = 0; r < rows; r++) {
 		g_print ("\tRow %02d -------\n", r);
 		for (c = 0; c < cols; c++) {
-			GdaValue *value;
+			const GdaValue *value;
 			gchar *strvalue;
 			const gchar *colname;
 
 			colname = gda_data_model_get_column_title (model, c);
-			value = (GdaValue *) gda_data_model_get_value_at (model, c, r);
+			value = gda_data_model_get_value_at (model, c, r);
 			strvalue = gda_value_stringify (value);
-			g_print ("\t\tColumn %02d name: %s\n", 
-					c, colname);
+			g_print ("\t\tColumn %02d name: %s\n", c, colname);
 			g_print ("\t\t         value: ('%s')\n", strvalue);
 			g_free (strvalue);
 		}
 	}
+
+	/* Do it again. Just to measure the speed */
+	for (r = 0; r < rows; r++) {
+		for (c = 0; c < cols; c++) {
+			gda_data_model_get_column_title (model, c);
+		}
+	}
+}
+
+void
+test_speed_random (GdaDataModel *model)
+{
+	gint cols, rows;
+	gint i, c, r;
+	GRand *generator;
+	GTimer *timer;
+
+	cols = gda_data_model_get_n_columns (model);
+	rows = gda_data_model_get_n_rows (model);
+
+	g_print (" Testing access to values of data model %p, with %d columns and %d rows\n",
+		 model, cols, rows);
+
+	generator = g_rand_new ();
+	timer = g_timer_new ();
+	for (i = 0; i < 10000; i++) {
+		c = g_rand_int_range (generator, 0, cols);
+		r = g_rand_int_range (generator, 0, rows);
+		(void) gda_data_model_get_value_at (model, c, r);
+	}
+	g_timer_stop (timer);
+	g_print (" Elapsed time for fetching 10000 random values: %f\n",
+		 g_timer_elapsed (timer, NULL));
+	g_timer_destroy (timer);
+	g_rand_free (generator);
 }
 
 void
@@ -113,10 +147,10 @@ display_data_model (GdaDataModel *model)
 
 	for (r = 0; r < rows; r++) {
 		for (c = 0; c < cols; c++) {
-			GdaValue *value;
+			const GdaValue *value;
 			gchar *str;
 
-			value = (GdaValue *) gda_data_model_get_value_at (model, c, r);
+			value = gda_data_model_get_value_at (model, c, r);
 			str = gda_value_stringify (value);
 			g_print ("\tColumn %d - Row %d: %s\n", c, r, str);
 			g_free (str);
@@ -189,7 +223,7 @@ test_array_model (void)
 		values = g_list_append (values, gda_value_new_string ("Another string"));
 		values = g_list_append (values, gda_value_new_double (4560.45672113323093));
 
-		gda_data_model_array_append_row (model, values);
+		gda_data_model_append_row (GDA_DATA_MODEL (model), values);
 
 		g_list_foreach (values, (GFunc) gda_value_free, NULL);
 		g_list_free (values);
